import re
from io import IOBase
from pathlib import Path
from subprocess import PIPE
from typing import AnyStr, IO, Union
from warnings import warn

import hatanaka.bin
from hatanaka import cli

__version__ = '1.0.0'
rnxcmp_version = '4.0.8'
__all__ = ['rnx2crx', 'crx2rnx', 'rnx2crx_file', 'crx2rnx_file', 'HatanakaException',
           'rnxcmp_version']


def rnx2crx(rnx_content: Union[AnyStr, IO], reinit_every_nth: int = None,
            skip_strange: bool = False) -> AnyStr:
    """Compress a RINEX observation file into the Compact RINEX format.

    Parameters
    ----------
    rnx_content : str or bytes or file-like
        RINEX observation file content
    reinit_every_nth : int, optional
        Initialize the compression operation at every # epochs.
        When some part of the Compact RINEX file is lost, the data can not be recovered
        thereafter until all the data arc are initialized for differential operation.
        This option may be used to increase chances to recover parts of data by using the
        skip_strange option of crx2rnx at the cost of increase of file size.
    skip_strange : bool, default False
        Warn and skip strange epochs instead of raising an exception.

    Returns
    -------
    str or bytes
        Compressed RINEX file content. bytes if rnx_content was binary, otherwise str.

    Raises
    ------
    HatanakaException
        On any errors during compression.

    Warns
    -----
    Any non-critical problems during compression will be raised as warnings.
    """
    extra_args = []
    if reinit_every_nth:
        assert isinstance(reinit_every_nth, int)
        extra_args += ['-e', '{:d}'.format(reinit_every_nth)]
    if skip_strange:
        extra_args += ['-s']
    return _run('rnx2crx', rnx_content, extra_args)


def crx2rnx(crx_content: Union[AnyStr, IO], skip_strange: bool = False) -> AnyStr:
    """Restore the original RINEX observation file from a Compact RINEX file.

    Parameters
    ----------
    crx_content : str or bytes or file-like
        Compact RINEX observation file content
    skip_strange : bool, default False
        Warn and skip strange epochs instead of raising an exception.
        This option may be used for salvaging usable data when middle of the Compact
        RINEX file is missing. The data after the missing part, are, however, useless
        until the compression operation of all data are initialized at some epoch.
        Using this together with of reinit_every_nth option of rnx2crx may be effective.
        Caution: It is assumed that no change in the list of data types happens in the
        lost part of the data.

    Returns
    -------
    str or bytes
        Decompressed RINEX file content. bytes if crx_content was binary, otherwise str.

    Raises
    ------
    HatanakaException
        On any errors during decompression.

    Warns
    -----
    Any non-critical problems during decompression will be raised as warnings.
    """
    extra_args = []
    if skip_strange:
        extra_args = ['-s']
    return _run('crx2rnx', crx_content, extra_args)


def rnx2crx_file(rnx_file: Union[str, Path], reinit_every_nth: int = None,
                 skip_strange: bool = False, overwrite: bool = False,
                 delete_on_success: bool = False) -> str:
    """Compress a RINEX observation file into the Compact RINEX format.

    Parameters
    ----------
    rnx_file : str or Path
        RINEX observation file path
    reinit_every_nth : int, optional
        Initialize the compression operation at every # epochs.
        When some part of the Compact RINEX file is lost, the data can not be recovered
        thereafter until all the data arc are initialized for differential operation.
        This option may be used to increase chances to recover parts of data by using the
        skip_strange option of crx2rnx at the cost of increase of file size.
    skip_strange : bool, default False
        Warn and skip strange epochs instead of raising an exception.
    overwrite : bool, default False
        Allow the output file to overwrite an existing one.
    delete_on_success : bool, default False
        Delete the input file if conversion finishes without errors and warnings.

    Returns
    -------
    str
        Path of the resulting compressed RINEX file.

    Raises
    ------
    HatanakaException
        On any errors during compression.

    Warns
    -----
    Any non-critical problems during compression will be raised as warnings.
    """
    args = []
    if reinit_every_nth:
        assert isinstance(reinit_every_nth, int)
        args += ['-e', '{:d}'.format(reinit_every_nth)]
    if skip_strange:
        args.append('-s')
    if overwrite:
        args.append('-f')
    if delete_on_success:
        args.append('-d')
    _run_file('rnx2crx', rnx_file, args)

    # Construct the expect output file name
    suffix = Path(rnx_file).suffix
    if suffix[3] == 'o':
        suffix = suffix[:3] + 'd' + suffix[4:]
    elif suffix[3] == 'O':
        suffix = suffix[:3] + 'D' + suffix[4:]
    elif suffix == '.rnx':
        suffix = '.crx'
    elif suffix == '.RNX':
        suffix = '.CRX'
    out_file = Path(rnx_file).with_suffix(suffix)
    print(suffix, rnx_file, out_file)
    assert out_file != rnx_file
    assert out_file.exists()
    return str(out_file)


def crx2rnx_file(crx_file: Union[str, Path], skip_strange: bool = False, overwrite: bool = False,
                 delete_on_success: bool = False) -> str:
    """Restore the original RINEX observation file from a Compact RINEX file.

    Parameters
    ----------
    crx_file : str or bytes or file-like
        Compact RINEX observation file path
    skip_strange : bool, default False
        Warn and skip strange epochs instead of raising an exception.
        This option may be used for salvaging usable data when middle of the Compact
        RINEX file is missing. The data after the missing part, are, however, useless
        until the compression operation of all data are initialized at some epoch.
        Using this together with of reinit_every_nth option of rnx2crx may be effective.
        Caution: It is assumed that no change in the list of data types happens in the
        lost part of the data.
    overwrite : bool, default False
        Allow the output file to overwrite an existing one.
    delete_on_success : bool, default False
        Delete the input file if conversion finishes without errors and warnings.

    Returns
    -------
    str
        Path of the resulting decompressed RINEX file.

    Raises
    ------
    HatanakaException
        On any errors during decompression.

    Warns
    -----
    Any non-critical problems during decompression will be raised as warnings.
    """
    args = []
    if skip_strange:
        args.append('-s')
    if overwrite:
        args.append('-f')
    if delete_on_success:
        args.append('-d')

    _run_file('crx2rnx', crx_file, args)

    # Construct the expect output file name
    suffix = Path(crx_file).suffix
    if suffix[3] == 'd':
        suffix = suffix[:3] + 'o' + suffix[4:]
    elif suffix[3] == 'D':
        suffix = suffix[:3] + 'O' + suffix[4:]
    elif suffix == '.crx':
        suffix = '.rnx'
    elif suffix == '.CRX':
        suffix = '.RNX'
    out_file = Path(crx_file).with_suffix(suffix)
    assert out_file != crx_file
    assert out_file.exists()
    return str(out_file)


class HatanakaException(RuntimeError):
    pass


def _is_binary(f: IO) -> bool:
    return isinstance(f.read(0), bytes)


def _run(program, content, extra_args=[]):
    encoding = None
    errors = None
    if isinstance(content, IOBase):
        if not _is_binary(content):
            encoding = 'ascii'
            # let's be relaxed about non-ascii symbols as long as it decodes successfully
            errors = 'ignore'
        proc = cli._popen(program, ['-'] + extra_args,
                          stdout=PIPE, stderr=PIPE, stdin=content, encoding=encoding, errors=errors)
        stdout, stderr = proc.communicate()
    else:
        if isinstance(content, str):
            encoding = 'ascii'
            errors = 'ignore'
        proc = cli._popen(program, ['-'] + extra_args,
                          stdout=PIPE, stderr=PIPE, stdin=PIPE, encoding=encoding, errors=errors)
        stdout, stderr = proc.communicate(content)
    retcode = proc.poll()

    _check(program, retcode, stderr)
    return stdout


def _run_file(program, file, args):
    if not Path(file).exists():
        raise FileNotFoundError(f"'{file}' does not exist")
    if not Path(file).is_file():
        raise FileNotFoundError(f"'{file}' is not a file")
    proc = cli._popen(program, [str(file)] + args, stdout=PIPE, stderr=PIPE, stdin=PIPE)
    stdout, stderr = proc.communicate()
    retcode = proc.poll()
    _check(program, retcode, stderr)


def _check(program, retcode, stderr):
    """Raise HatanakaException on errors and report warnings"""
    if isinstance(stderr, bytes):
        stderr = stderr.decode('ascii', errors='backslashreplace').strip()
    if retcode not in (0, 2):
        stderr = re.sub('^ +', '', stderr, flags=re.M)
        stderr = re.sub('\n(?!WARNING|ERROR)', ' ', stderr)
        stderr = re.sub('^ERROR *: *', '', stderr, flags=re.M)
        raise HatanakaException(stderr)
    if retcode == 2 and not stderr:
        warn(f'{program}: exited with an unspecified warning')
    if stderr:
        stderr = re.sub('^ +', '', stderr.strip(), flags=re.M)
        stderr = stderr.replace('\n', ' ')
        stderr = re.sub('^ *WARNING *:? *', '\n', stderr, flags=re.M)
        stderr = re.sub('^\n', '', stderr)
        warn(f'{program}: {stderr}')



import platform
import subprocess
import sys
from typing import List, Optional

from importlib_resources import path

import hatanaka.bin

__all__ = ['rnx2crx', 'crx2rnx']


def rnx2crx(args: Optional[List[str]] = None) -> int:
    return _run('rnx2crx', args)


def crx2rnx(args: Optional[List[str]] = None) -> int:
    return _run('crx2rnx', args)


def _run(program, args=None):
    if args is None:
        args = sys.argv[1:]
    p = _popen(program, args)
    p.wait()
    return p.returncode


def _popen(program, args, **kwargs):
    if platform.system() == 'Windows':
        program += '.exe'
    with path(hatanaka.bin, program) as executable:
        return subprocess.Popen([str(executable)] + args, **kwargs)



import re
import shutil

import pytest
from importlib_resources import files

import hatanaka.test
from hatanaka import HatanakaException, crx2rnx, crx2rnx_file, rnx2crx, rnx2crx_file
from hatanaka.cli import crx2rnx as crx2rnx_cli, rnx2crx as rnx2crx_cli

crx_sample = files(hatanaka.test) / 'sample.crx'
rnx_sample = files(hatanaka.test) / 'sample.rnx'


@pytest.fixture
def crx_str():
    with crx_sample.open('r', encoding='ascii') as f:
        return f.read()


@pytest.fixture
def rnx_str():
    with rnx_sample.open('r', encoding='ascii') as f:
        return f.read()


@pytest.fixture
def crx_str_stream():
    with crx_sample.open('r', encoding='ascii') as f:
        yield f


@pytest.fixture
def rnx_str_stream():
    with rnx_sample.open('r', encoding='ascii') as f:
        yield f


@pytest.fixture
def crx_bytes():
    with crx_sample.open('rb') as f:
        return f.read()


@pytest.fixture
def rnx_bytes():
    with rnx_sample.open('rb') as f:
        return f.read()


@pytest.fixture
def crx_bytes_stream():
    with crx_sample.open('rb') as f:
        yield f


@pytest.fixture
def rnx_bytes_stream():
    with rnx_sample.open('rb') as f:
        yield f


def clean(txt):
    # Remove rnx2crx version number and timestamp for testing
    pattern = '^RNX2CRX.+CRINEX PROG / DATE$'
    if isinstance(txt, bytes):
        return re.sub(pattern.encode(), b'', txt.replace(b'\r', b''), flags=re.M)
    return re.sub(pattern, '', txt, flags=re.M)


def test_rnx2crx_str(rnx_str, crx_str):
    assert clean(rnx2crx(rnx_str)) == clean(crx_str)


def test_rnx2crx_bytes(rnx_bytes, crx_bytes):
    assert clean(rnx2crx(rnx_bytes)) == clean(crx_bytes)


def test_crx2rnx_str(crx_str, rnx_str):
    assert crx2rnx(crx_str) == rnx_str


def test_crx2rnx_bytes(crx_bytes, rnx_bytes):
    assert clean(crx2rnx(crx_bytes)) == clean(rnx_bytes)


def test_rnx2crx_str_stream(rnx_str_stream, crx_str):
    assert clean(rnx2crx(rnx_str_stream)) == clean(crx_str)


def test_rnx2crx_bytes_stream(rnx_bytes_stream, crx_bytes):
    assert clean(rnx2crx(rnx_bytes_stream)) == clean(crx_bytes)


def test_crx2rnx_str_stream(crx_str_stream, rnx_str):
    assert crx2rnx(crx_str_stream) == rnx_str


def test_crx2rnx_bytes_stream(crx_bytes_stream, rnx_bytes):
    assert clean(crx2rnx(crx_bytes_stream)) == clean(rnx_bytes)


def test_invalid():
    with pytest.raises(HatanakaException) as excinfo:
        crx2rnx('blah')
    msg = excinfo.value.args[0]
    assert msg.startswith('The file seems to be truncated in the middle.')


def test_warning(rnx_bytes, crx_bytes):
    rnx_bytes += b'\0\0\0'
    with pytest.warns(UserWarning) as record:
        converted = rnx2crx(rnx_bytes)
    assert clean(converted) == clean(crx_bytes)
    assert len(record) == 1
    assert record[0].message.args[0].startswith('rnx2crx: null characters')


def test_rnx2crx_extra_args_good(rnx_str, crx_str):
    converted = rnx2crx(rnx_str, reinit_every_nth=1, skip_strange=True)
    assert clean(converted) == clean(crx_str)


def test_rnx2crx_extra_args_warning(rnx_str, crx_str):
    rnx_str = rnx_str.replace('R19 129262004.57708', 'G13 130321269.80108')
    with pytest.warns(UserWarning) as record:
        converted = rnx2crx(rnx_str, reinit_every_nth=1, skip_strange=True)
    assert len(record) == 1
    assert (record[0].message.args[0] ==
            'rnx2crx: Duplicated satellite in one epoch at line 15. ... skip')
    # Only the header remains
    assert clean(crx_str).startswith(clean(converted))


def test_crx2rnx_extra_args_good(rnx_str, crx_str):
    converted = crx2rnx(crx_str, skip_strange=True)
    assert clean(converted) == clean(rnx_str)


def test_rnx2crx_cli(tmp_path, crx_str):
    sample_path = tmp_path / rnx_sample.name
    shutil.copy(rnx_sample, sample_path)
    rnx2crx_cli([str(sample_path)])
    expected_path = tmp_path / (rnx_sample.stem + '.crx')
    assert expected_path.exists()
    with expected_path.open() as f:
        result = f.read()
    assert clean(result) == clean(crx_str)
    shutil.rmtree(tmp_path)


def test_crx2rnx_cli(tmp_path, rnx_str):
    sample_path = tmp_path / crx_sample.name
    shutil.copy(crx_sample, sample_path)
    crx2rnx_cli([str(sample_path)])
    expected_path = tmp_path / (crx_sample.stem + '.rnx')
    assert expected_path.exists()
    with expected_path.open() as f:
        result = f.read()
    assert clean(result) == clean(rnx_str)
    shutil.rmtree(tmp_path)


def test_non_ascii_is_tolerated(rnx_bytes, crx_bytes):
    def add_non_ascii(txt):
        return txt.replace(
            b'VERSION / TYPE',
            b'VERSION / TYPE' + 'õäü'.encode().ljust(60) + b'COMMENT\n')

    converted = crx2rnx(add_non_ascii(crx_bytes))
    assert clean(converted) == clean(add_non_ascii(rnx_bytes))
    converted = rnx2crx(add_non_ascii(rnx_bytes))
    assert clean(converted) == clean(add_non_ascii(crx_bytes))


def test_stderr_bad_encoding(crx_bytes):
    """The errors and warnings can include cited garbage from invalid files
    and we should handle that gracefully."""
    with pytest.raises(HatanakaException) as excinfo:
        crx2rnx(crx_bytes[:-1] + b'\xFF\0')
    msg = excinfo.value.args[0]
    assert msg.startswith('The file seems to be truncated')
    assert msg.endswith('\\xff<end')


@pytest.mark.parametrize(
    'input_name, expected_output_name',
    [
        ('sample.rnx', 'sample.crx'),
        ('sample.RNX', 'sample.CRX'),
        ('sample.21o', 'sample.21d'),
        ('sample.21O', 'sample.21D'),
    ]
)
def test_rnx2crx_file(tmp_path, crx_str, input_name, expected_output_name):
    sample_path = tmp_path / input_name
    shutil.copy(rnx_sample, sample_path)
    converted_path = rnx2crx_file(sample_path)
    expected_path = tmp_path / expected_output_name
    assert converted_path == str(expected_path)
    assert expected_path.exists()
    with expected_path.open() as f:
        result = f.read()
    assert clean(result) == clean(crx_str)
    shutil.rmtree(tmp_path)


@pytest.mark.parametrize(
    'input_name, expected_output_name',
    [
        ('sample.crx', 'sample.rnx'),
        ('sample.CRX', 'sample.RNX'),
        ('sample.21d', 'sample.21o'),
        ('sample.21D', 'sample.21O'),
    ]
)
def test_crx2rnx_file(tmp_path, rnx_str, input_name, expected_output_name):
    sample_path = tmp_path / input_name
    shutil.copy(crx_sample, sample_path)
    converted_path = crx2rnx_file(sample_path)
    expected_path = tmp_path / expected_output_name
    assert converted_path == str(expected_path)
    assert expected_path.exists()
    with expected_path.open() as f:
        result = f.read()
    assert clean(result) == clean(rnx_str)
    shutil.rmtree(tmp_path)


if __name__ == '__main__':
    pytest.main()



import os
import shutil
import subprocess
from pathlib import Path

from setuptools import setup
from setuptools.command.build_clib import build_clib as _build_clib


class build_clib(_build_clib):
    def build_libraries(self, libraries):
        cc = os.environ.get('CC')
        build_static = os.environ.get('BUILD_STATIC', 'n') == 'y'
        include_dirs = None
        library_dirs = None
        if cc is None:
            if self.compiler.compiler_type == 'msvc':
                self.compiler.initialize()
                cc = self.compiler.cc
                include_dirs = self.compiler.include_dirs
                library_dirs = self.compiler.library_dirs
            elif self.compiler.compiler_type == 'bcpp':
                cc = 'bcpp'
            elif hasattr(self.compiler, 'compiler_so') and self.compiler.compiler_so:
                # looks like ['gcc', '-pthread', '-Wl,--sysroot=/', ...]
                cc = self.compiler.compiler_so[0]
            if not cc:
                cc = find_c_compiler()
        output_dir = Path('build/lib/hatanaka/bin')
        output_dir.mkdir(parents=True, exist_ok=True)
        for executable, build_info in libraries:
            output = output_dir / executable
            build(build_info['sources'], output, cc, build_static, include_dirs, library_dirs)

        # copy to source dir as well for easier testing
        for f in list(output_dir.glob('rnx2crx*')) + list(output_dir.glob('crx2rnx*')):
            shutil.copy(f, 'hatanaka/bin/')


def build(sources, output, cc, build_static=False, include_dirs=None, library_dirs=None):
    if not all(Path(src).is_file() for src in sources):
        raise FileNotFoundError(sources)
    output = str(output)

    if cc.replace('.exe', '').endswith('cl'):  # msvc-like
        cmd = [cc, *sources, '/nologo', '/O2', '/Fe:' + output]
        if include_dirs:
            cmd += ['/I' + inc_dir for inc_dir in include_dirs]
        if library_dirs:
            cmd += ['/link']
            cmd += ['/LIBPATH:' + library_dir for library_dir in library_dirs]
    else:
        cmd = [cc, *sources, '-O3', '-Wno-unused-result', '-o', output]
        if include_dirs:
            cmd += ['-I' + inc_dir for inc_dir in include_dirs]
        if library_dirs:
            cmd += ['-L' + library_dir for library_dir in library_dirs]
        if build_static:
            cmd.append('-static')

    print(' '.join(cmd))
    subprocess.check_call(cmd)


def find_c_compiler(cc=None):
    compilers = ['cc', 'gcc', 'clang', 'icc', 'icl', 'cl', 'clang-cl']
    if cc is not None:
        compilers = [cc] + compilers
    available = list(filter(shutil.which, compilers))
    if not available:
        raise FileNotFoundError('No C compiler found on PATH')
    return available[0]


cmdclass = {'build_clib': build_clib}

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel


    class bdist_wheel(_bdist_wheel):
        def get_tag(self):
            impl, abi_tag, plat_name = super().get_tag()
            impl = 'py3'
            abi_tag = 'none'
            plat_name = plat_name.replace('linux', 'manylinux1')
            return impl, abi_tag, plat_name


    cmdclass['bdist_wheel'] = bdist_wheel
except ImportError:
    pass

setup(
    libraries=[
        ('rnx2crx', {'sources': ['rnxcmp/source/rnx2crx.c']}),
        ('crx2rnx', {'sources': ['rnxcmp/source/crx2rnx.c']})
    ],
    cmdclass=cmdclass,
)
